@@ -57,7 +57,7 @@ typedef struct {
     _cstack *cs;
     _htab *rec_levels;
     long id;
-    char *name;
+    PyObject *name;
     long long t0;                           // profiling start CPU time
     unsigned long sched_cnt;                // how many times this thread is scheduled
 } _ctx; // context
@@ -82,8 +82,9 @@ static int paused;
 static time_t yappstarttime;
 static long long yappstarttick;
 static long long yappstoptick;
-static _ctx *prev_ctx;
-static _ctx *current_ctx;
+static _ctx *prev_ctx = NULL;
+static _ctx *current_ctx = NULL;
+static _ctx *initial_ctx = NULL; // used for holding the context that called start()
 static PyObject *context_id_callback = NULL;
 static PyObject *context_name_callback = NULL;
 static PyObject *test_timings; // used for testing
@@ -169,57 +170,40 @@ _create_ctx(void)
     return ctx;
 }
 
-char *
-_current_ctx_name(void)
+static PyObject *
+_current_context_name(void)
 {
-    char *rc = NULL;
-    PyObject *callback_rc = NULL;
-    PyObject *mthr, *cthr, *tattr1, *tattr2;
-    mthr = cthr = tattr1 = tattr2 = NULL;
+    PyObject *name;
 
-    if (context_name_callback) {
-        callback_rc = PyObject_CallFunctionObjArgs(context_name_callback, NULL);
-        if (!callback_rc) {
-            PyErr_Print();
-            goto err;
-        }
-        if (!PyStr_Check(callback_rc)) {
-            yerr("context name callback returned non-string");
-            goto err;
-        }
-
-        rc = PyStr_AS_CSTRING(callback_rc);
-        Py_CLEAR(callback_rc);
-
-        return rc;
-    } else {
-        mthr = PyImport_ImportModuleNoBlock("threading"); // Requires Python 2.6.
-        if (!mthr)
-            goto err;
-        cthr = PyObject_CallMethod(mthr, "currentThread", "");
-        if (!cthr)
-            goto err;
-        tattr1 = PyObject_GetAttrString(cthr, "__class__");
-        if (!tattr1)
-            goto err;
-        tattr2 = PyObject_GetAttrString(tattr1, "__name__");
-        if (!tattr2)
-            goto err;
-
-        Py_DECREF(mthr);
-        Py_DECREF(cthr);
-        Py_DECREF(tattr1);
-        Py_DECREF(tattr2);
-        return PyStr_AS_CSTRING(tattr2);
+    if (!context_name_callback) {
+        return NULL;
     }
+
+    name = PyObject_CallFunctionObjArgs(context_name_callback, NULL);
+    if (!name) {
+        PyErr_Print();
+        goto err;
+    }
+
+    if (name == Py_None) {
+        // Name not available yet - will try again on the next call
+        goto later;
+    }
+
+    if (!PyStr_Check(name)) {
+        yerr("context name callback returned non-string");
+        goto err;
+    }
+
+    return name;
+
 err:
     PyErr_Clear();
-    Py_XDECREF(mthr);
-    Py_XDECREF(cthr);
-    Py_XDECREF(tattr1);
-    Py_XDECREF(tattr2);
-    Py_XDECREF(callback_rc);
     Py_CLEAR(context_name_callback);  /* Don't use the callback again. */
+    Py_XDECREF(name);
+    return NULL;
+later:
+    Py_XDECREF(name);
     return NULL;
 }
 
@@ -252,9 +236,7 @@ _current_context_id(PyThreadState *ts)
 
 error:
     PyErr_Clear();
-    // don't use callback again
-    Py_XDECREF(context_id_callback);
-    context_id_callback = NULL;
+    Py_CLEAR(context_id_callback); // don't use callback again
     return 0;
 }
 
@@ -285,8 +267,41 @@ _del_pit(_pit *pit)
         it = next;
     }
     pit->children = NULL;
-    Py_XDECREF(pit->name);
-    Py_XDECREF(pit->modname);
+    Py_CLEAR(pit->name);
+    Py_CLEAR(pit->modname);
+}
+
+static PyObject *
+_pycfunction_module_name(PyCFunctionObject *cfn)
+{
+    PyObject *obj;
+    PyObject *name;
+
+    // The __module__ attribute, can be anything
+    obj = cfn->m_module;
+
+    if (!obj) {
+        // TODO: Is this always correct?
+        name = PyStr_FromString("__builtin__");
+    } else if (PyStr_Check(obj)) {
+        Py_INCREF(obj);
+        name = obj;
+    } else if (PyModule_Check(obj)) {
+        const char *s = PyModule_GetName(obj);
+        if (!s) {
+            goto error;
+        }
+        name = PyStr_FromString(s);
+    } else {
+        // Something else - str(obj)
+        name = PyObject_Str(obj);
+    }
+
+    return name;
+
+error:
+    PyErr_Clear();
+    return PyStr_FromString("<unknown>");
 }
 
 static _pit *
@@ -294,8 +309,6 @@ _ccode2pit(void *cco)
 {
     PyCFunctionObject *cfn;
     _hitem *it;
-    PyObject *mod;
-    char *modname;
     PyObject *name;
 
     cfn = cco;
@@ -312,25 +325,7 @@ _ccode2pit(void *cco)
             return NULL;
 
         pit->builtin = 1;
-
-        // get module name
-        modname = NULL;
-        mod = cfn->m_module;
-
-        if (mod) {
-            if (PyStr_Check(mod)) {
-                modname = PyStr_AS_CSTRING(mod);
-            } else if (PyModule_Check(mod)) {
-                modname = (char *)PyModule_GetName(mod);
-            }
-        }
-
-        if (modname == NULL) {
-            PyErr_Clear();
-            modname = "__builtin__";
-        }
-
-        pit->modname = PyStr_FromString(modname);
+        pit->modname = _pycfunction_module_name(cfn);
         pit->lineno = 0;
 
         // built-in method?
@@ -377,7 +372,8 @@ _code2pit(PyFrameObject *fobj)
         return NULL;
 
     pit->name = NULL;
-    pit->modname = PyStr_FromString(PyStr_AS_CSTRING(cobj->co_filename));
+    Py_INCREF(cobj->co_filename);
+    pit->modname = cobj->co_filename;
     pit->lineno = cobj->co_firstlineno;
 
     PyFrame_FastToLocals(fobj);
@@ -403,7 +399,8 @@ _code2pit(PyFrameObject *fobj)
         }
     }
     if (!pit->name) {
-        pit->name = PyStr_FromString(PyStr_AS_CSTRING(cobj->co_name));
+        Py_INCREF(cobj->co_name);
+        pit->name = cobj->co_name;
     }
 
     PyFrame_LocalsToFast(fobj, 0);
@@ -678,6 +675,7 @@ _del_ctx(_ctx * ctx)
 {
     sdestroy(ctx->cs);
     htdestroy(ctx->rec_levels);
+    Py_CLEAR(ctx->name);
 }
 
 static int
@@ -695,6 +693,12 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
         goto finally;
     }
 
+    // do not profile if multi-threaded is off and the context is different than
+    // the context that called start.
+    if (!flags.multithreaded && current_ctx != initial_ctx) {
+        goto finally;
+    }
+
     // update ctx stats
     if (prev_ctx != current_ctx) {
         current_ctx->sched_cnt++;
@@ -702,7 +706,7 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
     prev_ctx = current_ctx;
     if (!current_ctx->name)
     {
-        current_ctx->name = _current_ctx_name();
+        current_ctx->name = _current_context_name();
     }
 
 
@@ -785,13 +789,8 @@ _unprofile_thread(PyThreadState *ts)
 static void
 _ensure_thread_profiled(PyThreadState *ts)
 {
-    PyThreadState *p = NULL;
-
-    for (p=ts->interp->tstate_head ; p != NULL; p = p->next) {
-        if (ts->c_profilefunc != _yapp_callback) {
-            _profile_thread(ts);
-        }
-    }
+    if (ts->c_profilefunc != _yapp_callback)
+        _profile_thread(ts);
 }
 
 static void
@@ -811,22 +810,39 @@ _init_profiler(void)
     if (!yappinitialized) {
         contexts = htcreate(HT_CTX_SIZE);
         if (!contexts)
-            return 0;
+            goto error;
         pits = htcreate(HT_PIT_SIZE);
         if (!pits)
-            return 0;
+            goto error;
         flpit = flcreate(sizeof(_pit), FL_PIT_SIZE);
         if (!flpit)
-            return 0;
+            goto error;
         flctx = flcreate(sizeof(_ctx), FL_CTX_SIZE);
         if (!flctx)
-            return 0;
+            goto error;
         yappinitialized = 1;
-        current_ctx = NULL;
-        prev_ctx = NULL;
-        ycurfuncindex = 0;
     }
     return 1;
+
+error:
+    if (contexts) {
+        htdestroy(contexts);
+        contexts = NULL;
+    }
+    if (pits) {
+        htdestroy(pits);
+        pits = NULL;
+    }
+    if (flpit) {
+        fldestroy(flpit);
+        flpit = NULL;
+    }
+    if (flctx) {
+        fldestroy(flctx);
+        flctx = NULL;
+    }
+
+    return 0;
 }
 
 static PyObject*
@@ -902,6 +918,7 @@ _start(void)
         _enum_threads(&_profile_thread);
     } else {
         _ensure_thread_profiled(PyThreadState_GET());
+        initial_ctx = _thread2ctx(PyThreadState_GET());
     }
 
     yapprunning = 1;
@@ -930,17 +947,28 @@ clear_stats(PyObject *self, PyObject *args)
         Py_RETURN_NONE;
     }
 
+    current_ctx = NULL;
+    prev_ctx = NULL;
+    initial_ctx = NULL;
+
     henum(pits, _pitenumdel, NULL);
     htdestroy(pits);
+    pits = NULL;
+
     henum(contexts, _ctxenumdel, NULL);
     htdestroy(contexts);
+    contexts = NULL;
 
     fldestroy(flpit);
+    flpit = NULL;
+
     fldestroy(flctx);
+    flctx = NULL;
+
     yappinitialized = 0;
     yapphavestats = 0;
-    Py_XDECREF(test_timings);
-    test_timings = NULL;
+    ycurfuncindex = 0;
+    Py_CLEAR(test_timings);
 
 // check for mem leaks if DEBUG_MEM is specified
 #ifdef DEBUG_MEM
@@ -964,7 +992,7 @@ static int
 _ctxenumstat(_hitem *item, void *arg)
 {
     PyObject *efn;
-    char *tcname;
+    const char *tcname;
     _ctx *ctx;
     long long cumdiff;
     PyObject *exc;
@@ -981,9 +1009,11 @@ _ctxenumstat(_hitem *item, void *arg)
         return 0;
     }
 
-    tcname = ctx->name;
-    if (tcname == NULL)
+    if (ctx->name)
+        tcname = PyStr_AS_CSTRING(ctx->name);
+    else
         tcname = UNINITIALIZED_STRING_VAL;
+
     efn = (PyObject *)arg;
 
     cumdiff = _calc_cumdiff(tickcount(), ctx->t0);
