@@ -10,6 +10,10 @@ import _yappi
 import pickle
 import marshal
 import threading
+try:
+    from thread import get_ident        # Python 2
+except ImportError:
+    from threading import get_ident     # Python 3
 from collections import defaultdict
 
 class YappiError(Exception): pass
@@ -53,19 +57,34 @@ def _validate_columns(name, list):
         raise YappiError("Invalid Column name: '%s'" % (name))
 
 
-"""
- _callback will only be called once per-thread. _yappi will detect
- the new thread and changes the profilefunc param of the ThreadState
- structure. This is an internal function please don't mess with it.
-"""
-def _callback(frame, event, arg):
-    _yappi._profile_event(frame, event, arg)
-    return _callback
+def _ctx_name_callback():
+    """
+    We don't use threading.current_thread() becuase it will deadlock if
+    called when profiling threading._active_limbo_lock.acquire().
+    See: #Issue48.
+    """
+    try:
+        current_thread = threading._active[get_ident()]
+        return current_thread.__class__.__name__
+    except KeyError: 
+        # Threads may not be registered yet in first few profile callbacks.
+        return None
 
-"""
-function to prettify time columns in stats.
-"""
+# set _ctx_name_callback by default at import time. 
+_yappi.set_context_name_callback(_ctx_name_callback)
+
+def _profile_thread_callback(frame, event, arg):
+    """
+    _profile_thread_callback will only be called once per-thread. _yappi will detect
+    the new thread and changes the profilefunc param of the ThreadState
+    structure. This is an internal function please don't mess with it.
+    """
+    _yappi._profile_event(frame, event, arg)
+    
 def _fft(x, COL_SIZE=8):
+    """
+    function to prettify time columns in stats.
+    """
     _rprecision = 6
     while(_rprecision > 0):
         _fmt = "%0." + "%d" % (_rprecision) + "f"
@@ -759,7 +778,7 @@ def start(builtins=False, profile_threads=True):
     Start profiler.
     """
     if profile_threads:
-        threading.setprofile(_callback)
+        threading.setprofile(_profile_thread_callback)
     _yappi.start(builtins, profile_threads)
 
 def get_func_stats():
@@ -861,14 +880,16 @@ def set_context_id_callback(callback):
 
 def set_context_name_callback(callback):
     """
-    Use a string other than the current thread's class name as the context's
-    name.
+    Set the callback to retrieve current context's name.
 
     The callback must take no arguments and return a string. For example:
 
     >>> import greenlet, yappi
     >>> yappi.set_context_name_callback(
     ...     lambda: greenlet.getcurrent().__class__.__name__)
+
+    If the callback cannot return the name at this time but may be able to
+    return it later, it should return None.
     """
     return _yappi.set_context_name_callback(callback)
 
